// Project Sword & Gun Copyright © 2021 Kevin Poretti


#include "Characters/SNGCharacterBase.h"

#include "DrawDebugHelpers.h"
#include "Camera/CameraComponent.h"
#include "Components/CapsuleComponent.h"
#include "Components/InputComponent.h"
#include "Components/SNGCharacterMovementComponent.h"
#include "GameFramework/SpringArmComponent.h"
#include "Kismet/KismetMathLibrary.h"

static int32 CharacterMovementDebug = 0;
FAutoConsoleVariableRef CVARCharacterMovementDebug(TEXT("SNG.CharacterMovementDebug.Show"),
						CharacterMovementDebug, 
						TEXT("Print movement information for character"), 
						ECVF_Cheat);

static int32 ShowCharacterTrajectory = 0;
FAutoConsoleVariableRef CVARShowCharacterTrajectory(TEXT("SNG.CharacterMovementDebug.Trajectory.Show"),
						ShowCharacterTrajectory, 
						TEXT("Render character trajectory"), 
						ECVF_Cheat);

static float TrajectoryPeriod = 0.05f;
FAutoConsoleVariableRef CVARTrajectoryPeriod(TEXT("SNG.CharacterMovementDebug.Trajectory.Period"),
						TrajectoryPeriod, 
						TEXT("How often should trajectory lines be drawn"), 
						ECVF_Cheat);

static float TrajectoryLineLifetime = 10.0f;
FAutoConsoleVariableRef CVARTrajectoryLineLifetime(TEXT("SNG.CharacterMovementDebug.Trajectory.Lifetime"),
						TrajectoryLineLifetime, 
						TEXT("How long should a trajectory line segment be visible"), 
						ECVF_Cheat);

static float LastSavedPosTime = 0.0f;
static FVector LastSavedPos = FVector::ZeroVector;

// Sets default values
ASNGCharacterBase::ASNGCharacterBase(const FObjectInitializer& ObjectInitializer)
	: Super(ObjectInitializer.SetDefaultSubobjectClass<USNGCharacterMovementComponent>(ACharacter::CharacterMovementComponentName))
{
 	// Set this character to call Tick() every frame.  You can turn this off to improve performance if you don't need it.
	PrimaryActorTick.bCanEverTick = true;

	SpringArmComponent = CreateDefaultSubobject<USpringArmComponent>(TEXT("SpringArmComponent"));
	SpringArmComponent->SetupAttachment(RootComponent);
	SpringArmComponent->bUsePawnControlRotation = true;
	SpringArmComponent->bInheritYaw = true;
	SpringArmComponent->bInheritPitch = true;
	SpringArmComponent->bInheritRoll = true;
	
	CameraComponent = CreateDefaultSubobject<UCameraComponent>(TEXT("CameraComponent"));
	CameraComponent->SetupAttachment(SpringArmComponent);
	
	bUseControllerRotationYaw = false;
	bUseControllerRotationPitch = false;
	bUseControllerRotationRoll = false;
	
	USkeletalMeshComponent* CharMesh = GetMesh();
	if(CharMesh)
	{
		CharMesh->SetRelativeLocation(FVector(0.0f, 0.0f, -GetCapsuleComponent()->GetScaledCapsuleHalfHeight()));
		CharMesh->SetRelativeRotation(FRotator(0.0f, -90.0f, 0.0f));
	}
	else
	{
		UE_LOG(LogTemp, Warning, TEXT("SNGCharacterBase :: Could not get skeletal mesh. Did you forget to set it?"));
	}

	InputThreshold = 0.1f;

	DefaultFOV = 90.0f;
	SprintFOV = 110.0f;
	SprintFOVInterpSpeed = 2.0f;
}

void ASNGCharacterBase::MoveForward(float Value)
{
	if(!CanMove)
	{
		return;
	}
	
	InputDirection.Y = Value;

	UAnimMontage* CurrMontage = GetCurrentMontage();
	if(CurrMontage && !InputDirection.IsNearlyZero(0.1) && !GetCharacterMovement()->IsFalling() && CanMovementCancelMontage)
	{
		StopAnimMontage(CurrMontage);
	}
	
	FRotator ViewRotation = FRotator(0.0f, GetControlRotation().Yaw, 0.0f);
	AddMovementInput(UKismetMathLibrary::GetForwardVector(ViewRotation), Value);
}

void ASNGCharacterBase::MoveRight(float Value)
{
	if(!CanMove)
	{
		return;
	}
	
	InputDirection.X = Value;

	UAnimMontage* CurrMontage = GetCurrentMontage();
	if(CurrMontage && !InputDirection.IsNearlyZero(0.1) && !GetCharacterMovement()->IsFalling() && CanMovementCancelMontage)
	{
		StopAnimMontage(CurrMontage);
	}
	
	FRotator ViewRotation = FRotator(0.0f, GetControlRotation().Yaw, 0.0f);
	AddMovementInput(UKismetMathLibrary::GetRightVector(ViewRotation), Value);
}

void ASNGCharacterBase::TurnRate(float Value)
{
	AddControllerYawInput(Value * 50.0f * GetWorld()->DeltaTimeSeconds);
}

void ASNGCharacterBase::LookUpRate(float Value)
{
	AddControllerPitchInput(Value * 50.0f * GetWorld()->DeltaTimeSeconds);
}

void ASNGCharacterBase::StartJump()
{
	if(!CanJump())
	{
		return;
	}

	// can't rotate in air
	GetCharacterMovement()->bOrientRotationToMovement = false;

	CanMovementCancelMontage = false;
	
	FVector ActorDir = GetActorForwardVector();
	FVector InputDir = GetInputAsWorldDirection();
	float Dot = UKismetMathLibrary::Dot_VectorVector(ActorDir, InputDir);
	if(Dot > 0.5f)
	{
		check(JumpFwdMontage);
		PlayAnimMontage(JumpFwdMontage);
	}
	else if (Dot < -0.5f)
	{
		check(JumpBackMontage);
		PlayAnimMontage(JumpBackMontage);
	}
	else
	{
		check(JumpNeutralMontage);
		PlayAnimMontage(JumpNeutralMontage);
	}
}

void ASNGCharacterBase::Landed(const FHitResult& Hit)
{
	Super::Landed(Hit);

	// can rotate when on the ground again
	GetCharacterMovement()->bOrientRotationToMovement = true;
	CanMovementCancelMontage = true;
	
	PlayAnimMontage(JumpNeutralMontage, 1.0f, "Land");
}

// Called when the game starts or when spawned
void ASNGCharacterBase::BeginPlay()
{
	Super::BeginPlay();

	CanMovementCancelMontage = false;
	CanMove = true;
}

FVector ASNGCharacterBase::GetInputAsWorldDirection()
{
	AController* MyController = GetController();
	if(MyController == nullptr)
	{
		return FVector::ZeroVector;
	}
	
	FRotator ViewRotation = FRotator(0.0f, MyController->GetControlRotation().Yaw, 0.0f);
	FVector ForwardInput = UKismetMathLibrary::GetForwardVector(ViewRotation) * InputDirection.Y;
	FVector RightInput = UKismetMathLibrary::GetRightVector(ViewRotation) * InputDirection.X;

	return ForwardInput + RightInput;
}

void ASNGCharacterBase::ApplyHitStop(float StopTime, float SlowTime)
{
	UE_LOG(LogTemp, Error, TEXT("HIT STOP IS NOT IMPLEMENTED"));
}

// Called every frame
void ASNGCharacterBase::Tick(float DeltaTime)
{
	Super::Tick(DeltaTime);

	USNGCharacterMovementComponent* CharMovement = Cast<USNGCharacterMovementComponent>(GetCharacterMovement());
	check(CharMovement);
	// interpolate to appropriate field of view depending on whether player sprinting or not
	float TargetFOV = CharMovement->GetIsSprinting() ? SprintFOV : DefaultFOV;
	float NewFOV = FMath::FInterpTo(CameraComponent->FieldOfView, TargetFOV, DeltaTime, SprintFOVInterpSpeed);
	CameraComponent->SetFieldOfView(NewFOV);
	
	// Debug
	if(CharacterMovementDebug)
	{
		float LateralSpeed = GetVelocity().Size2D();
		FVector VelocityDir = GetVelocity();
		VelocityDir.Normalize();
		FVector LateralVelocityDir = FVector(VelocityDir.X, VelocityDir.Y, 0.0f); 
		
		FVector StartPos(GetActorLocation().X,
						 GetActorLocation().Y,
						GetActorLocation().Z - GetCapsuleComponent()->GetScaledCapsuleHalfHeight());
		FVector ActorForward = UKismetMathLibrary::GetForwardVector(GetActorRotation());

		// Draw arrow showing velocity 
		DrawDebugDirectionalArrow(GetWorld(),
			StartPos, StartPos + (VelocityDir * ((GetVelocity().Size() / GetCharacterMovement()->MaxWalkSpeed) * 100.0f)),
			10.0f, FColor::Yellow, false, -1.0f, 0, 2.0f);
		
		// Draw arrow showing lateral velocity
		StartPos.Z += 1; // avoid z-fighting
		DrawDebugDirectionalArrow(GetWorld(),
			StartPos, StartPos + (LateralVelocityDir * ((LateralSpeed / GetCharacterMovement()->MaxWalkSpeed) * 100.0f)),
			10.0f, FColor::Orange, false, -1.0f, 0, 2.0f);
		
		// draw arrow showing what direction the player is inputting (WASD or right thumbstick) in the world relative to the camera 
		FVector CurrInputDirection = GetInputAsWorldDirection();
		CurrInputDirection.Normalize();
		StartPos.Z += 1; // avoid z-fighting
		DrawDebugDirectionalArrow(GetWorld(), StartPos, StartPos + (CurrInputDirection * 100.0f),
		10.0f, FColor::Purple, false, -1.0f, 0, 2.0f);

		// Draw arrow showing where character is facing
		StartPos.Z += 1; // avoid z-fighting
		DrawDebugDirectionalArrow(GetWorld(),
			StartPos, StartPos + (ActorForward * 100.0f),
			10.0f, FColor::Blue, false, -1.0f, 0, 2.0f);
		
		FString DebugMsg = FString::Printf(TEXT("MaxWalkSpeed: %f"), GetCharacterMovement()->MaxWalkSpeed);
		GEngine->AddOnScreenDebugMessage(-1, 0.0f, FColor::Magenta, DebugMsg);
		
		DebugMsg = FString::Printf(TEXT("LateralSpeed: %f"), LateralSpeed);
		GEngine->AddOnScreenDebugMessage(-1, 0.0f, FColor::Magenta, DebugMsg);
		
		DebugMsg = FString::Printf(TEXT("RotationRate: %s"), *GetCharacterMovement()->RotationRate.ToString());
		GEngine->AddOnScreenDebugMessage(-1, 0.0f, FColor::Magenta, DebugMsg);

		bool InputThresholdIsMet = CurrInputDirection.Size2D() >= InputThreshold;
		DebugMsg = FString::Printf(TEXT("InputThresholdIsMet: %hs"), InputThresholdIsMet ? "true" : "false");
		GEngine->AddOnScreenDebugMessage(-1, 0.0f, FColor::Magenta, DebugMsg);
		
		if(ShowCharacterTrajectory)
		{
			if(GetWorld()->TimeSeconds - LastSavedPosTime >= TrajectoryPeriod)
			{
				DrawDebugLine(GetWorld(), LastSavedPos, StartPos, FColor::Red, false, TrajectoryLineLifetime, 0, 2.0f);
				LastSavedPos = StartPos;
				LastSavedPosTime = GetWorld()->TimeSeconds;
			}
		}
	}
}

// Called to bind functionality to input
void ASNGCharacterBase::SetupPlayerInputComponent(UInputComponent* PlayerInputComponent)
{
	Super::SetupPlayerInputComponent(PlayerInputComponent);

	// char movement
    PlayerInputComponent->BindAxis("MoveForward", this, &ASNGCharacterBase::MoveForward);
	PlayerInputComponent->BindAxis("MoveRight", this, &ASNGCharacterBase::MoveRight);

	// camera movement
	PlayerInputComponent->BindAxis("Turn", this, &ASNGCharacterBase::AddControllerYawInput);
	PlayerInputComponent->BindAxis("LookUp", this, &ASNGCharacterBase::AddControllerPitchInput);

	// controller camera movement
	PlayerInputComponent->BindAxis("TurnRate", this, &ASNGCharacterBase::TurnRate);
	PlayerInputComponent->BindAxis("LookUpRate", this, &ASNGCharacterBase::LookUpRate);

	// actions
	PlayerInputComponent->BindAction("Jump", IE_Pressed, this, &ASNGCharacterBase::StartJump);
	//PlayerInputComponent->BindAction("Jump", IE_Released, this, &ASNGCharacterBase::StopJumping);
}

