// Project Sword & Gun Copyright © 2021 Kevin Poretti


#include "Components/SNGCharacterMovementComponent.h"

#include "Characters/SNGCharacterBase.h"
#include "GameFramework/Character.h"
#include "Kismet/KismetMathLibrary.h"
#include "PhysicsEngine/PhysicsSettings.h"

USNGCharacterMovementComponent::USNGCharacterMovementComponent()
{
	// Defaults
	DefaultGroundFriction = 8.0f;
	DefaultMaxWalkSpeed = 600.0f;
	DefaultRotationRate = FRotator(0.0f, 360.0f, 0.0f);
	
	// sprint
	SprintingMaxWalkSpeed = 1000.0f;
	SprintLateralSpeedThreshold = 500.0f;
	SprintingRotationRate = FRotator(0.0f, 180.0f, 0.0f);
	TimeToInitiateSprint = 2.0f;
	WalkToSprintInterpSpeed = 4.0f;

	// firing
	FiringMaxWalkSpeed = 300.0f;
	
	// reloading
	ReloadingMaxWalkSpeed = 300.0f;
	
	// airborne defaults
	DefaultAirControl = 0.3f;
	
	// jump
	MaxJumpHeight = 250.0f;
	DesiredJumpDuration = 0.75f;
	DefaultJumpLateralVelocity = DefaultMaxWalkSpeed;
	LateralVelocityScaling = FVector2D(1.0f, 1.0f);
	ComputeGravityScaleAndJumpZVelocity();

	// dash
	DashVelocity = 1000.0f;
	DashGroundFriction = 0.0f;
	DashGravityScale = 0.1f;
	DashFloatTime = 0.2f;
	DashSlideTime = 0.5f;
	DashGravityToDefaultSpeed = 5.0f;
	DashGroundFrictionToDefaultSpeed = 5.0f;
	
	/*
	// not used...yet
	MinJumpHeight = 200.0f;
	MaxJumpTime = 0.2f;

	LateralVelocityDampening = 0.8f;
	
	MinAllowedDashHeight = 175.0f;
	MinAllowedAttackHeight = 175.0f;
	*/

	// Character Movement component defaults
	SetWalkableFloorAngle(45.622231f);
	
	AirControlBoostMultiplier = 0.0f;
	AirControlBoostVelocityThreshold = 0.0f;
}

/**
 *	Instantly sets all movement parameters to their defaults values (i.e. ground friction to DefaultGroundFriction,
 *	MaxWalkSpeed to DefaultMaxWalkSpeed, etc.)
 */
void USNGCharacterMovementComponent::RestoreMovementDefaults()
{
	GroundFriction = TargetGroundFriction = DefaultGroundFriction;
	GravityScale = TargetGravityScale = DefaultGravityScale;
	RotationRate = DefaultRotationRate;
	MaxWalkSpeed = TargetMaxWalkSpeed = DefaultMaxWalkSpeed;
	AirControl = DefaultAirControl;
	JumpZVelocity = DefaultJumpZVelocity;
}

void USNGCharacterMovementComponent::Dash()
{
	FLaunchParameters DashParams;
	DashParams.bOverrideLateralVelocityClamping = true;
	DashParams.bOverrideX = true;
	DashParams.bOverrideY = true;
	DashParams.bOverrideZ = true;
	DashParams.bOverrideMovementMode = false;
	DashParams.GroundFriction = DashGroundFriction;
	DashParams.GravityScale = DashGravityScale;
	DashParams.FloatTime = DashFloatTime;
	DashParams.SlideTime = DashSlideTime;
	
	// sets gravity and friction targets which will start interpolating after float time and slide time elapse respectively
	TargetGravityScale = DefaultGravityScale;
	TargetGroundFriction = DefaultGroundFriction;
	GravityScaleInterpSpeed = DashGravityToDefaultSpeed;
	GroundFrictionInterpSpeed = DashGroundFrictionToDefaultSpeed;
	
	FRotator CharacterRot = FRotator(0.0f, CharacterOwner->GetActorRotation().Yaw, 0.0f);
	FVector LaunchDir = FVector(UKismetMathLibrary::GetForwardVector(CharacterRot));
	LaunchDir.Normalize();

	DashParams.Velocity = LaunchDir * DashVelocity;
	LaunchCharacter(DashParams);
}

#if WITH_EDITOR
void USNGCharacterMovementComponent::PostEditChangeProperty(FPropertyChangedEvent& PropertyChangedEvent)
{
	Super::PostEditChangeProperty(PropertyChangedEvent);

	const FProperty* PropertyThatChanged = PropertyChangedEvent.MemberProperty;
	if (PropertyThatChanged && PropertyThatChanged->GetFName() == GET_MEMBER_NAME_CHECKED(USNGCharacterMovementComponent, MaxJumpHeight) ||
		PropertyThatChanged && PropertyThatChanged->GetFName() == GET_MEMBER_NAME_CHECKED(USNGCharacterMovementComponent, DesiredJumpDuration))
	{
		// Compute WalkableFloorZ from the Angle.
		ComputeGravityScaleAndJumpZVelocity();
	}
}

void USNGCharacterMovementComponent::PostLoad()
{
	Super::PostLoad();

	// Compute gravity scale and jump z velocity for first time
	ComputeGravityScaleAndJumpZVelocity();
}
#endif // WITH_EDITOR

void USNGCharacterMovementComponent::OnSprintTimer()
{
	bIsSprinting = true;
	TargetMaxWalkSpeed = SprintingMaxWalkSpeed;
}

void USNGCharacterMovementComponent::LaunchCharacter(const FLaunchParameters LaunchParams)
{
	Velocity *= LaunchParams.VelocityScale;
	
	if(LaunchParams.bOverrideX)
	{
		Velocity.X = LaunchParams.Velocity.X;
	}
	else
	{
		Velocity.X += LaunchParams.Velocity.X;
	}

	if(LaunchParams.bOverrideY)
	{
		Velocity.Y = LaunchParams.Velocity.Y;
	}
	else
	{
		Velocity.Y += LaunchParams.Velocity.Y;
	}
	
	if(LaunchParams.bOverrideZ)
	{
		Velocity.Z = LaunchParams.Velocity.Z;
	}
	else
	{
		Velocity.Z += LaunchParams.Velocity.Z;
	}

	if(!LaunchParams.bOverrideLateralVelocityClamping)
	{
		// clamp lateral velocity to max walk speed
		FVector LateralVelocity = FVector(Velocity.X, Velocity.Y, 0.0f);
		LateralVelocity = LateralVelocity.GetClampedToMaxSize(DefaultMaxWalkSpeed);
		Velocity = FVector(LateralVelocity.X, LateralVelocity.Y, Velocity.Z);
	}
	
	GroundFriction = LaunchParams.bUseDefaultGroundFriction ? DefaultGroundFriction : LaunchParams.GroundFriction;
	GravityScale = LaunchParams.bUseDefaultGravityScale ? DefaultGravityScale : LaunchParams.GravityScale;
	FloatTime = LaunchParams.FloatTime;
	SlideTime = LaunchParams.SlideTime;

	if(LaunchParams.bOverrideMovementMode)
	{
		SetMovementMode(LaunchParams.NewMovementMode);
	}
}

void USNGCharacterMovementComponent::ComputeGravityScaleAndJumpZVelocity()
{
	float HalfJumpDuration = DesiredJumpDuration / 2.0f;
	DefaultJumpZVelocity = (2 * MaxJumpHeight) / HalfJumpDuration;
	
	float GameGravity = UPhysicsSettings::Get()->DefaultGravityZ;
	float TargetGravity = (-2 * MaxJumpHeight) / (HalfJumpDuration * HalfJumpDuration);
	DefaultGravityScale = TargetGravity / GameGravity;
}

bool USNGCharacterMovementComponent::DoJump(bool bReplayingMoves)
{
	if ( CharacterOwner && CharacterOwner->CanJump() )
	{
		// Don't jump if we can't move up/down.
		if (!bConstrainToPlane || FMath::Abs(PlaneConstraintNormal.Z) != 1.f)
		{
			ASNGCharacterBase* CharacterOwnerAsSNGChar = Cast<ASNGCharacterBase>(CharacterOwner);
			if(CharacterOwnerAsSNGChar)
			{
				FLaunchParameters JumpParams;

				FVector WorldInputDir = CharacterOwnerAsSNGChar->GetInputAsWorldDirection();
				bool InputThresholdMet = WorldInputDir.Size2D() >= CharacterOwnerAsSNGChar->GetInputThreshold();
				if(InputThresholdMet)
				{
					FVector JumpLateralVelocity = WorldInputDir * DefaultJumpLateralVelocity;
					JumpParams.Velocity.X = JumpLateralVelocity.X;
					JumpParams.Velocity.Y = JumpLateralVelocity.Y;
				}
				
				JumpParams.Velocity.Z = DefaultJumpZVelocity;
				JumpParams.bOverrideLateralVelocityClamping = false;
				JumpParams.bOverrideX = false;
				JumpParams.bOverrideY = false;
				JumpParams.bOverrideZ = true;

				/**
				 *	NOTE(kevin): This is done so that if the player holds nothing the character still carries some lateral
				 *	momentum out of a neutral jump. If the player if moving forward but jumps right (this is all relative
				 *	to the camera) we want all forward momentum canceled and the character to only have velocity going
				 *	up and right.
				 */
				JumpParams.VelocityScale = FVector(InputThresholdMet ? FVector2D(WorldInputDir.X, WorldInputDir.Y) * LateralVelocityScaling : LateralVelocityScaling, 1.0f);
				
				JumpParams.bUseDefaultGroundFriction = true;
				JumpParams.bUseDefaultGravityScale = true;
			
				JumpParams.bOverrideMovementMode = true;
				JumpParams.NewMovementMode = EMovementMode::MOVE_Falling;

				LaunchCharacter(JumpParams);
				return true;
			}
			else
			{
				UE_LOG(LogTemp, Error, TEXT("SNGCharacterMovementComponent :: DoJump :: Could not cast CharacterOwner to ASNGCharacterBase"));
				return false;
			}
		}
	}
	
	return false;
}

void USNGCharacterMovementComponent::InterpolateMovementParameters(float DeltaTime)
{
	// set walk speed
	MaxWalkSpeed = FMath::FInterpTo(MaxWalkSpeed, TargetMaxWalkSpeed, DeltaTime, MaxWalkSpeedInterpSpeed);

	// ground friction
	if(SlideTime > 0.0f)
	{
		SlideTime = FMath::Max(SlideTime - DeltaTime * GetCharacterOwner()->CustomTimeDilation, 0.0f);
	}
	else
	{
		GroundFriction = FMath::FInterpTo(GroundFriction, TargetGroundFriction, DeltaTime, GroundFrictionInterpSpeed);
	}

	// gravity scale
	if(FloatTime > 0.0f)
	{
		FloatTime = FMath::Max(FloatTime - DeltaTime * GetCharacterOwner()->CustomTimeDilation, 0.0f);
	}
	else
	{
		GravityScale = FMath::FInterpTo(GravityScale, TargetGravityScale, DeltaTime, GravityScaleInterpSpeed);
	}
	
	// set rotation rate
	// NOTE(kevin): since default rotate rate is 360 RInterpTo will go from 0 to 180 instead of 360 to 180
	// need to think of a way to prevent this
	// might just need to do my own interp manually
	/*
	FRotator TargetRotationRate = bIsSprinting ? SprintingRotationRate : DefaultRotationRate;
	FRotator NewRotationRate = FMath::RInterpTo(RotationRate, TargetRotationRate, DeltaTime, WalkToSprintInterpSpeed);
	RotationRate = NewRotationRate;
	*/

	// gravity
}

void USNGCharacterMovementComponent::BeginPlay()
{
	Super::BeginPlay();

	// set character movement parameters to defaults
	RestoreMovementDefaults();
}

void USNGCharacterMovementComponent::TickComponent(float DeltaTime, ELevelTick Tick,
	FActorComponentTickFunction* ThisTickFunction)
{
	Super::TickComponent(DeltaTime, Tick, ThisTickFunction);

	// turning off sprinting for now...not a HUGE fan of this implementation
	/*
	float LateralSpeed = Velocity.Size2D();
	if(LateralSpeed >= SprintLateralSpeedThreshold && !GetWorld()->GetTimerManager().IsTimerActive(TimerHandle_SprintTimer))
	{
		GetWorld()->GetTimerManager().SetTimer(TimerHandle_SprintTimer,
			this, &USNGCharacterMovementComponent::OnSprintTimer,
			TimeToInitiateSprint, false);
	}
	
	if(LateralSpeed < SprintLateralSpeedThreshold)
	{
		if(GetWorld()->GetTimerManager().IsTimerActive(TimerHandle_SprintTimer))
		{
			GetWorld()->GetTimerManager().ClearTimer(TimerHandle_SprintTimer);
		}
		bIsSprinting = false;
		TargetMaxWalkSpeed = DefaultMaxWalkSpeed;
		MaxWalkSpeedInterpSpeed = WalkToSprintInterpSpeed;
	}
	*/
	
	InterpolateMovementParameters(DeltaTime);
}
