// Project Sword & Gun Copyright © 2021 Kevin Poretti


#include "States/SNGState.h"

USNGState::USNGState()
{
	Name = NAME_None;
	TimeInState = 0.0f;
	bCanTransitionToItself = false;
	DefaultSubStateIndex = 0;
}

void USNGState::RegisterTransition(FName DestState, FName TransitionRuleFunctionName)
{
	
	FStateTransitionRule StateTransitionRule;
	StateTransitionRule.BindUFunction(this, TransitionRuleFunctionName);
	
	if(!Transitions.Contains(DestState))
	{
		Transitions.Add(DestState, StateTransitionRule);
	}
	else
	{
		UE_LOG(LogTemp, Warning, TEXT("USNGState :: Transition dictionary already contains transition rule for destination state %s"), *DestState.ToString());
	}
}

void USNGState::RegisterEventHandler(ESNGEventType EventType, FName EventHandlerFunctionName)
{
	FEventHandlerRule EventHandlerRule;
	EventHandlerRule.BindUFunction(this, EventHandlerFunctionName);

	if(Handlers.Contains(EventType))
	{
		Handlers[EventType].Add(EventHandlerRule);
	}
	else
	{
		Handlers.Add(EventType, TArray<FEventHandlerRule>());
		Handlers[EventType].Add(EventHandlerRule);
	}
}

FName USNGState::GetNextStateFromEvent(FSNGEvent& Event)
{
	FName DestState = NAME_None;
	if(Handlers.Contains(Event.EventType))
	{
		for (auto EventHandler : Handlers[Event.EventType])
		{
			// FStateTransitionRule is a delegate that returns true if the transition rules are met
			// and false otherwise
			if(!EventHandler.IsBound())
			{
				UE_LOG(LogTemp, Error, TEXT("SNGState :: Tried calling function %s but it is not bound"), *EventHandler.GetFunctionName().ToString());
				continue;
			}
			
			DestState = EventHandler.Execute(Event);
			if(DestState != NAME_None)
			{
				return DestState;
			}
		}
	}

	if(Parent)
	{
		return Parent->GetNextStateFromEvent(Event);
	}
	
	return NAME_None;
}

FName USNGState::GetNextState()
{
	// NOTE(kevin): Does this actually have to be a map?
	for(const TPair<FName, FStateTransitionRule>& pair : Transitions)
	{
		// FStateTransitionRule is a delegate that returns true if the transition rules are met
		// and false otherwise
		if(!pair.Value.IsBound())
		{
			UE_LOG(LogTemp, Error, TEXT("SNGState :: Tried calling function %s but it is not bound"), *pair.Value.GetFunctionName().ToString());
			continue;
		}
		
		if(pair.Value.Execute())
		{
			return pair.Key;
		}
	}

	if(Parent)
	{
		return Parent->GetNextState();
	}
	
	return NAME_None;
}

AActor* USNGState::GetOwner()
{
	return Owner;
}

void USNGState::SetOwner(AActor* StateOwner)
{
	Owner = StateOwner;
}

USNGState* USNGState::GetParent()
{
	return Parent;
}

void USNGState::SetParent(USNGState* ParentState)
{
	this->Parent = ParentState;
}

uint32 USNGState::GetDepth()
{
	return Depth;
}

void USNGState::CreateState()
{
	// pre-compute the  depth/level of this state during creation
	// used later for calculating least common ancestor of two states
	uint32 CurrDepth = 0;
	USNGState* CurrState = this;
	while(CurrState->Parent != nullptr)
	{
		CurrDepth++;
		CurrState = CurrState->Parent;
	}
	Depth = CurrDepth;
	
	OnCreateState();
}

USNGState* USNGState::EnterState(FSNGEvent& Event)
{
	OnEnterState(Event);
	TimeInState = 0.0f;

	if(SubStates.Num() > 0 &&
		(DefaultSubStateIndex >= 0 && DefaultSubStateIndex < SubStates.Num()))
	{
		return SubStates[DefaultSubStateIndex]->EnterState(Event);	
	}
	return this;
}

void USNGState::ExitState()
{
	OnExitState();

	// to do this correctly i think we need to find the shared parent of the current state and the next state
	// we only want to exit states until that point
	if(Parent)
	{
		Parent->ExitState();
	}
}

void USNGState::UpdateState(float DeltaTime)
{
	OnUpdateState(DeltaTime);
	TimeInState += DeltaTime;

	if(Parent)
	{
		Parent->UpdateState(DeltaTime);
	}
}

FName USNGState::GetName()
{
	return Name;
}

float USNGState::GetTimeInState()
{
	return TimeInState;
}

bool USNGState::CanTransitionToItself()
{
	return bCanTransitionToItself;
}

TArray<TSubclassOf<USNGState>> USNGState::GetSubStateClasses()
{
	return SubStateClasses;
}

void USNGState::AddSubState(USNGState* SubState)
{
	SubStates.Add(SubState);
}
