// Project Sword & Gun Copyright © 2021 Kevin Poretti

#pragma once

#include "CoreMinimal.h"
#include "GameFramework/CharacterMovementComponent.h"
#include "SNGCharacterMovementComponent.generated.h"

USTRUCT(BlueprintType)
struct FLaunchParameters
{
	GENERATED_BODY()

	/** Velocity to either add or set for this character */
	UPROPERTY(EditAnywhere, BlueprintReadWrite)
	FVector Velocity;

	/**
	 *	How much to scale current velocity by. Set all components to 1 if you want the character to keep whatever
	 *	velocity they currently have.
	 */
	UPROPERTY(EditAnywhere, BlueprintReadWrite)
	FVector VelocityScale;

	/** If true, lateral velocity will NOT be clamped to max walk speed after performing all other velocity update operations */ 
	UPROPERTY(EditAnywhere, BlueprintReadWrite)
	bool bOverrideLateralVelocityClamping;
	
	/**
	 *	Should the X component of character's velocity be overriden with Velocity parameter or added?
	 *	Default is false which means Velocity X component will be added to current character's velocity
	 */
	UPROPERTY(EditAnywhere, BlueprintReadWrite)
	bool bOverrideX;

	/**
	 *	Should the Y component of character's velocity be overriden with Velocity parameter or added?
	 *	Default is false which means Velocity Y component will be added to current character's velocity
	 */
	UPROPERTY(EditAnywhere, BlueprintReadWrite)
	bool bOverrideY;

	/**
	 *	Should the Z component of character's velocity be overriden with Velocity parameter or added?
	 *	Default is false which means Velocity Z component will be added to current character's velocity
	 */
	UPROPERTY(EditAnywhere, BlueprintReadWrite)
	bool bOverrideZ;
	
	/** Should the character's movement mode change after being launched? */
	UPROPERTY(EditAnywhere, BlueprintReadWrite)
	bool bOverrideMovementMode;

	/** What the character's movement mode will be set to after launch. This will only be set is bOverrideMovementMode is true. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite)
	TEnumAsByte<EMovementMode> NewMovementMode;

	/**
	 *	GroundFriction will be overriden to DefaultGroundFriction as specified in the character movement
	 *	component instead of what is supplied in this struct. 
	 */
	UPROPERTY(EditAnywhere, BlueprintReadWrite)
	bool bUseDefaultGroundFriction;
	
	/** The ground friction that should be immediately applied when the character is launched */
	UPROPERTY(EditAnywhere, BlueprintReadWrite)
	float GroundFriction;

	/**
	 *	GravityScale will be overriden to DefaultGravityScale as specified in the character movement
	 *	component instead of what is supplied in this struct. 
	 */
	UPROPERTY(EditAnywhere, BlueprintReadWrite)
	bool bUseDefaultGravityScale;
	
	/** The gravity scale that should be immediately applied when the character is launched */
	UPROPERTY(EditAnywhere, BlueprintReadWrite)
	float GravityScale;
	
	/** The float time that should be immediately applied when the character is launched */
	UPROPERTY(EditAnywhere, BlueprintReadWrite)
	float FloatTime;

	/** The slide time that should be immediately applied when the character is launched */
	UPROPERTY(EditAnywhere, BlueprintReadWrite)
	float SlideTime;

	FLaunchParameters()
	: Velocity(FVector::ZeroVector)
	, VelocityScale(FVector::OneVector)
	, bOverrideLateralVelocityClamping(false)
	, bOverrideX(false)
	, bOverrideY(false)
	, bOverrideZ(false)
	, bOverrideMovementMode(false)
	, NewMovementMode(EMovementMode::MOVE_Walking)
	, bUseDefaultGroundFriction(false)
	, GroundFriction(8.0f)
	, bUseDefaultGravityScale(false)
	, GravityScale(1.0f)
	, FloatTime(0.0f)
	, SlideTime(0.0f)
	{
	}
};

/**
 *	Class that extends UE CharacterMovementComponent to add SNG specific functionality
 *
 *	Allows for different gravity scales, friction, and walk speeds when performing certain actions. This class will also
 *	lerp and changed values 
 */
UCLASS()
class SWORDNGUN_API USNGCharacterMovementComponent : public UCharacterMovementComponent
{
	GENERATED_BODY()

public:
	USNGCharacterMovementComponent();

	virtual void PostLoad() override;
	
	/**
	 *	Instantly sets all movement parameters to their defaults values (i.e. ground friction to DefaultGroundFriction,
	 *	MaxWalkSpeed to DefaultMaxWalkSpeed, etc.)
	 */
	UFUNCTION(BlueprintCallable)
	void RestoreMovementDefaults();

	/**
	 * @brief	Sets and scales current character's velocity and sets other movement values specified in the passed LaunchParams.
	 *			The order of operations for updating velocity is:
	 *			1) Multiply current velocity by VelocityScale
	 *			2) Will either set or add to current velocity components with the value specified in LaunchParams.Velocity 
	 * @param	LaunchParams - settings on how this character should be launched. See documentation for FLaunchParameters
	 *			for more information. 
	 */
	UFUNCTION(BlueprintCallable, meta=(AutoCreateRefTerm="LaunchParams"))
	void LaunchCharacter(const FLaunchParameters LaunchParams);

	/**
	 * @brief	Sets up launch parameters and character movement parameter targets and finally launches the character
	 *			for the character to dash.
	 */
	UFUNCTION(BlueprintCallable)
	void Dash();
	
	/**
	 *	Gets bool whether the character is currently in the sprinting state
	 */
	UFUNCTION(BlueprintCallable)
	FORCEINLINE bool GetIsSprinting() { return bIsSprinting; }
	
#if WITH_EDITOR
	virtual void PostEditChangeProperty(FPropertyChangedEvent& PropertyChangedEvent) override;
#endif // WITH_EDITOR

protected:
	// default ground 
	/** Default friction when not performing any special actions and on the ground */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Defaults|Ground", meta=(ClampMin="0", UIMin="0"))
	float DefaultGroundFriction;

	/** Ground friction we want character to interpolate to. Interpolated based on GroundFrictionInterpSpeed. */
	UPROPERTY(BlueprintReadWrite)
	float TargetGroundFriction;

	/** How fast current GroundFriction will be interpolated to TargetGroundFriction */
	UPROPERTY(BlueprintReadWrite)
	float GroundFrictionInterpSpeed;

	/** Timer used to determine when to start interpolating GroundFriction back to TargetGroundFriction */
	UPROPERTY(BlueprintReadWrite)
	float SlideTime;
	
	/** Default ground walk speed when not performing any special actions */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Defaults|Ground", meta=(ClampMin="0", UIMin="0"))
	float DefaultMaxWalkSpeed;

	/** Current MaxWalkSpeed we want to character to walk at. Interpolated based on MaxWalkSpeedInterpSpeed. */
	UPROPERTY(BlueprintReadWrite)
	float TargetMaxWalkSpeed;

	UPROPERTY(BlueprintReadWrite)
	/** How fast current MaxWalkSpeed will be interpolated to TargetMaxWalkSpeed */
	float MaxWalkSpeedInterpSpeed;
	
	/** Default ground walk speed when not performing any special actions */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Defaults|Ground", meta=(ClampMin="0", UIMin="0"))
	FRotator DefaultRotationRate;
	
	// sprinting 
	/** Ground walk speed when sprinting */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Sprint", meta=(ClampMin="0", UIMin="0"))
	float SprintingMaxWalkSpeed;

	/** Lateral speed the player needs to maintain to stay in the sprinting state */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Sprint", meta=(ClampMin="0", UIMin="0"))
	float SprintLateralSpeedThreshold;

	/** Default ground walk speed when not performing any special actions */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Sprint", meta=(ClampMin="0", UIMin="0"))
	FRotator SprintingRotationRate;
	
	/** How long after applying movement input should the character start sprinting? */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Sprint", meta=(ClampMin="0", UIMin="0"))
	float TimeToInitiateSprint;

	/** The speed in which the max walk speed to interpolated to the max sprint speed when a sprint has been initiated */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Sprint", meta=(ClampMin="0", UIMin="0"))
	float WalkToSprintInterpSpeed;
	
	FTimerHandle TimerHandle_SprintTimer;

	UPROPERTY(BlueprintReadOnly)
	bool bIsSprinting;
	
	void OnSprintTimer();

	// firing
	/** Ground walk speed when firing a ranged weapon */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Firing", meta=(ClampMin="0", UIMin="0"))
	float FiringMaxWalkSpeed;

	// reloading
	/** Ground walk speed when reloading a ranged weapon */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Reloading", meta=(ClampMin="0", UIMin="0"))
	float ReloadingMaxWalkSpeed;

	// airborne
	/** Gravity scale when not performing any special actions */
	UPROPERTY(VisibleAnywhere, BlueprintReadOnly, Category="SNG Character Movement|Defaults|Air", meta=(ClampMin="0", UIMin="0"))
	float DefaultGravityScale;

	/** Gravity Scale we want character to interpolate to. Interpolated based on GravityScaleInterpSpeed. */
	UPROPERTY(BlueprintReadWrite)
	float TargetGravityScale;

	/** How fast current GravityScale will be interpolated to TargetGravityScale */
	UPROPERTY(BlueprintReadWrite)
	float GravityScaleInterpSpeed;

	/** Timer used to determine when to start interpolating GravityScale back to TargetGravityScale */
	UPROPERTY(BlueprintReadWrite)
	float FloatTime;
	
	/**
	 *	How high (Z) is the character able to reach when holding the jump button for the maximum allowed time.
	 *
	 *	This is used in conjunction with DesiredJumpDuration to derive DefaultGravityScale and JumpZVelocity.
	 */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Jump", meta=(ClampMin="0", UIMin="0"))
	float MaxJumpHeight;

	/**
	 *	How long the character should be in the air for when performing a single jump and holding the jump button to reach
	 *	the maximum jump height.
	 *
	 *	This is used in conjunction with MaxJumpHeight to derive DefaultGravityScale and JumpZVelocity.
	 */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Jump", meta=(ClampMin="0", UIMin="0"))
	float DesiredJumpDuration;
	
	/** Z velocity set on the character when they are press the jump button */
	UPROPERTY(VisibleAnywhere, BlueprintReadOnly, Category="SNG Character Movement|Jump", meta=(ClampMin="0", UIMin="0"))
	float DefaultJumpZVelocity;
	
	/** Minimum height a character must reach before jump force is no longer applied even if the jump input is not being held */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Jump", meta=(ClampMin="0", UIMin="0"))
	float MinJumpHeight;

	/** Max time the jump button can be held to give character upward velocity */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Jump", meta=(ClampMin="0", UIMin="0"))
	float MaxJumpTime;

	/**
	 *	Lateral velocity to set when the player initiates jump and is holding a direction.
	 *
	 *	This value is multiplied by the world direction the player is inputting.
	 */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Jump", meta=(ClampMin="0", UIMin="0"))
	float DefaultJumpLateralVelocity;
	
	/** How much to scale lateral velocity by when a jump is initiated */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Jump", meta=(ClampMin="0", UIMin="0"))
	FVector2D LateralVelocityScaling;
	
	void ComputeGravityScaleAndJumpZVelocity();
	
	virtual bool DoJump(bool bReplayingMoves) override;

	/** Default friction when not performing any special actions and in the air */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Defaults|Air", meta=(ClampMin="0", UIMin="0"))
	float DefaultAirControl;

	/** Minimum distance from the ground a character has to be to be able to perform an air dash */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Defaults|Air", meta=(ClampMin="0", UIMin="0"))
	float MinAllowedDashHeight;

	/** Minimum distance from the ground a character has to be to be able to perform an air attack */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Defaults|Air", meta=(ClampMin="0", UIMin="0"))
	float MinAllowedAttackHeight;

	/** Initial lateral velocity when the character dashes */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Dash", meta=(ClampMin="0", UIMin="0"))
	float DashVelocity;

	/** Initial ground friction when the character dashes */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Dash", meta=(ClampMin="0", UIMin="0"))
	float DashGroundFriction;

	/** Initial gravity scale when the character dashes */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Dash", meta=(ClampMin="0", UIMin="0"))
	float DashGravityScale;

	/** How long the character should "float" before the gravity scale is interpolated back to default */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Dash", meta=(ClampMin="0", UIMin="0"))
	float DashFloatTime;

	/** How fast default gravity scale will be interpolated to after dash is done */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Dash", meta=(ClampMin="0", UIMin="0"))
	float DashGravityToDefaultSpeed;
	
	/** How long the character should "slide" before the ground friction is interpolated back to default */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Dash", meta=(ClampMin="0", UIMin="0"))
	float DashSlideTime;

	/** How fast default ground friction will be interpolated to after dash is done */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Dash", meta=(ClampMin="0", UIMin="0"))
	float DashGroundFrictionToDefaultSpeed;
	
	/**
	 * @brief Interpolates current movement parameters (i.e. GravityScale, MaxWalkSpeed, GroundFrinction) to the values
	 * specified by the parameters corresponding target value (i.e. GravityScale will be interpolated to TargetGravityScale).
	 * The interpolation speed is determined by the parameters corresponding speed value (i.e. GravityScale is interpolated
	 * to TargetGravityScale by a speed of GravityScaleInterpSpeed).
	 * 
	 * This function is called every component tick.
	 */
	void InterpolateMovementParameters(float DeltaTime);
	
	virtual void BeginPlay() override;

	virtual void TickComponent(float DeltaTime, ELevelTick TickType, FActorComponentTickFunction* ThisTickFunction) override;
};
